/*
 * Copyright © 2016-2017 Triplequote LLC. All rights reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Triplequote LLC. The intellectual and technical
 * concepts contained herein are proprietary to Triplequote LLC
 * and may be covered by Swiss and Foreign Patents, patents in process,
 * and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Triplequote LLC.
 */
/*
 * Zinc - The incremental compiler for Scala.
 * Copyright 2011 - 2017, Lightbend, Inc.
 * Copyright 2008 - 2010, Mark Harrah
 * This software is released under the terms written in LICENSE.
 *
 * This implementation contains changes made by Triplequote.
 */

package xsbt

import xsbti.{AnalysisCallback, Logger, Problem, Reporter}
import xsbti.compile._
import scala.tools.nsc.{Settings, reporters}

import Log.debug
import java.io.File

import com.triplequote.hydra.{Hydra, HydraSettings}

final class CompilerInterface {
  def newCompiler(options: Array[String],
                  output: Output,
                  initialLog: Logger,
                  initialDelegate: Reporter): CachedCompiler =
    new CachedCompiler0(options, output, new WeakLog(initialLog, initialDelegate))

  def run(sources: Array[File],
          changes: DependencyChanges,
          callback: AnalysisCallback,
          log: Logger,
          delegate: Reporter,
          progress: CompileProgress,
          cached: CachedCompiler): Unit =
    cached.run(sources, changes, callback, log, delegate, progress)
}

class InterfaceCompileFailed(val arguments: Array[String],
                             val problems: Array[Problem],
                             override val toString: String)
  extends xsbti.CompileFailed

class InterfaceCompileCancelled(val arguments: Array[String], override val toString: String)
  extends xsbti.CompileCancelled

private final class WeakLog(private[this] var log: Logger, private[this] var delegate: Reporter) {
  def apply(message: String): Unit = {
    assert(log ne null, "Stale reference to logger")
    log.error(Message(message))
  }
  def logger: Logger = log
  def reporter: Reporter = delegate
  def clear(): Unit = {
    log = null
    delegate = null
  }
}

private final class CachedCompiler0(args: Array[String], output: Output, initialLog: WeakLog)
  extends CachedCompiler
    with CachedCompilerCompat
    with java.io.Closeable {

  /////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////// INITIALIZATION CODE ////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////////////////////////

  private val command = {
    val settings = new HydraSettings(s => initialLog(s))
    // side-effect settings
    val outputDir = output match {
      case multi: MultipleOutput =>
        for (out <- multi.getOutputGroups)
          settings.outputDirs
            .add(out.getSourceDirectory.getAbsolutePath, out.getOutputDirectory.getAbsolutePath)
      case single: SingleOutput =>
        val outputFilepath = single.getOutputDirectory.getAbsolutePath
        settings.outputDirs.setSingleOutput(outputFilepath)
    }
    Command(args.toList, settings)
  }

  initialize(initialLog.logger)

  private[this] def initialize(logger: Logger): Unit = {
    try {
      val dreporter = DelegatingReporter(command.settings, initialLog.reporter)
      // Execute arguments' validation logic only once at instantation. If an argument is changed,
      // a new instance of this class is created by Sbt.
      Hydra.validateArguments(command) match {
        case Left(Hydra.CompilerSettingsError(msg)) =>
          dreporter.printSummary()
          handleErrors(dreporter, logger, msg)
        case Right(_) => ()
      }

      if (!noErrors(dreporter)) {
        dreporter.printSummary()
        handleErrors(dreporter, logger, "Compilation failed (CompilerInterface)")
      }
    }
    finally
      initialLog.clear()
  }

  /////////////////////////////////////////////////////////////////////////////////////////////////

  def close(): Unit = () // no-op because the ZincCompiler is not a field in our bridge implementation

  def noErrors(dreporter: DelegatingReporter) = !dreporter.hasErrors && command.ok

  def commandArguments(sources: Array[File]): Array[String] =
    (command.settings.recreateArgs ++ sources.map(_.getAbsolutePath)).toArray[String]

  import scala.tools.nsc.Properties.versionString
  def infoOnCachedCompiler(compilerId: String): String =
    s"[zinc] Running cached compiler $compilerId for Hydra Scala compiler $versionString"

  def run(sources: Array[File],
          changes: DependencyChanges,
          callback: AnalysisCallback,
          log: Logger,
          delegate: Reporter,
          progress: CompileProgress): Unit = synchronized {
    debug(log, infoOnCachedCompiler(hashCode().toLong.toHexString))
    debug(log, commandArguments(sources).mkString("Calling Hydra compiler with arguments  (CompilerInterface):\n\t", "\n\t", ""))
    val dreporter = DelegatingReporter(command.settings, delegate)
    try { run(sources.toList, changes, callback, log, dreporter, progress) } finally {
      dreporter.dropDelegate()
    }
  }

/*
  private def prettyPrintCompilationArguments(args: Array[String]) =
    args.mkString("[zinc] The Hydra Scala compiler is invoked with:\n\t", "\n\t", "")
*/

  private[this] def run(sources: List[File],
                        changes: DependencyChanges,
                        callback: AnalysisCallback,
                        log: Logger,
                        underlyingReporter: DelegatingReporter,
                        compileProgress: CompileProgress): Unit = {

    val newCompilerFactory = (reporter: reporters.Reporter) => {
      val compiler = newCompiler(command.settings, underlyingReporter, output)
      compiler.set(callback, underlyingReporter)
      compiler
    }
    val newRunFactory = (global: ZincCompiler) => new global.ZincRun(compileProgress)

    Hydra.compile(sources, command, newCompilerFactory, newRunFactory, underlyingReporter) match {
      case Left(Hydra.CompilerSettingsError(msg)) =>
        throw new InterfaceCompileFailed(args, Array(), msg)
      case Right(_) =>
        //processUnreportedWarnings(run)
        underlyingReporter.problems foreach { p =>
          callback.problem(p.category, p.position, p.message, p.severity, true)
        }
        underlyingReporter.printSummary()
    }

    if (!noErrors(underlyingReporter)) handleErrors(underlyingReporter, log, "Compilation failed (CompilerInterface)")
    // the case where we cancelled compilation _after_ some compilation errors got reported
    // will be handled by line above so errors still will be reported properly just potentially not
    // all of them (because we cancelled the compilation)
    if (underlyingReporter.cancelled) handleCompilationCancellation(underlyingReporter, log)
  }

  def handleErrors(dreporter: DelegatingReporter, log: Logger, msg: String): Nothing = {
    debug(log, msg)
    throw new InterfaceCompileFailed(args, dreporter.problems, msg)
  }

  def handleCompilationCancellation(dreporter: DelegatingReporter, log: Logger): Nothing = {
    assert(dreporter.cancelled, "We should get here only if when compilation got cancelled")
    debug(log, "Compilation cancelled (CompilerInterface)")
    throw new InterfaceCompileCancelled(args, "Compilation has been cancelled")
  }

/*
  def processUnreportedWarnings(run: compiler.Run): Unit = {
    // allConditionalWarnings and the ConditionalWarning class are only in 2.10+
    final class CondWarnCompat(val what: String,
                               val warnings: mutable.ListBuffer[(compiler.Position, String)])
    implicit def compat(run: AnyRef): Compat = new Compat
    final class Compat { def allConditionalWarnings = List[CondWarnCompat]() }

    val warnings = run.allConditionalWarnings
    if (warnings.nonEmpty)
      compiler.logUnreportedWarnings(warnings.map(cw => ("" /*cw.what*/, cw.warnings.toList)))
  }
*/
}
